# Step 1 Install required packages
# pip install pandas numpy matplotlib scikit-learn torch transformers datasets huggingface_hub seaborn

# Step 2: Import libraries
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler, OneHotEncoder

import torch
import torch.nn as nn
from torch.utils.data import Dataset, DataLoader

from transformers import PreTrainedTokenizerFast, AutoModelForSequenceClassification
from transformers import TrainingArguments, Trainer
from datasets import Dataset as HFDataset

from huggingface_hub import notebook_login

# Step 4: Generate or Load Data
df = pd.read_csv('city_temperature_data.csv')

# Step 5: Data Exploration and Preprocessing

# Display basic information
print("Dataset Shape:", df.shape)
print("\nDataset Info:")
print(df.info())
print("\nSummary Statistics:")
print(df.describe())

# Check for missing values
print("\nMissing Values:")
print(df.isnull().sum())

# Exploratory Data Analysis
''' PLOTS
# Plot temperature distribution
plt.figure(figsize=(12, 6))
sns.histplot(df['avg_temp'], bins=20, kde=True)
plt.title('Distribution of Average Temperatures')
plt.xlabel('Temperature (°C)')
plt.ylabel('Frequency')
plt.grid(True, alpha=0.3)
plt.show()

# Plot temperatures by city (boxplot)
plt.figure(figsize=(15, 8))
sns.boxplot(x='city', y='avg_temp', data=df)
plt.title('Average Temperature by City')
plt.xlabel('City')
plt.ylabel('Average Temperature (°C)')
plt.xticks(rotation=90)
plt.grid(True, alpha=0.3)
plt.tight_layout()
plt.show()

# Plot temperature by month
plt.figure(figsize=(12, 6))
monthly_avg = df.groupby('month')['avg_temp'].mean().reindex(
    ["January", "February", "March", "April", "May", "June", 
     "July", "August", "September", "October", "November", "December"]
)
sns.barplot(x=monthly_avg.index, y=monthly_avg.values)
plt.title('Global Average Temperature by Month')
plt.xlabel('Month')
plt.ylabel('Average Temperature (°C)')
plt.xticks(rotation=45)
plt.grid(True, alpha=0.3)
plt.tight_layout()
plt.show()
'''

# Feature Engineering

# Create month number feature (for cyclical encoding)
month_to_num = {month: i+1 for i, month in enumerate([
    "January", "February", "March", "April", "May", "June", 
    "July", "August", "September", "October", "November", "December"
])}
df['month_num'] = df['month'].map(month_to_num)

# Create cyclical features for month to capture seasonality
df['month_sin'] = np.sin(2 * np.pi * df['month_num']/12)
df['month_cos'] = np.cos(2 * np.pi * df['month_num']/12)

# Create text feature (will be used for tokenization)
df['text'] = "The temperature in " + df['city'] + " in " + df['month'] + " is"

# One-hot encode cities
city_encoder = OneHotEncoder(sparse_output=False)
city_encoded = city_encoder.fit_transform(df[['city']])
city_encoded_df = pd.DataFrame(
    city_encoded, 
    columns=[f'city_{city}' for city in city_encoder.categories_[0]]
)

# Standardize numerical features
scaler = StandardScaler()
numerical_features = ['month_sin', 'month_cos']
df[numerical_features] = scaler.fit_transform(df[numerical_features])

# Combine all prepared features
preprocessed_df = pd.concat([
    df[['text', 'city', 'month', 'month_num', 'month_sin', 'month_cos', 'avg_temp', 'high_temp', 'low_temp']],
    city_encoded_df
], axis=1)

# Display the preprocessed data
print("\nPreprocessed Data Sample:")
print(preprocessed_df.head())

# Split into train, validation, and test sets
train_df, temp_df = train_test_split(preprocessed_df, test_size=0.3, random_state=42)
val_df, test_df = train_test_split(temp_df, test_size=0.5, random_state=42)

print(f"\nTrain set: {train_df.shape[0]} samples")
print(f"Validation set: {val_df.shape[0]} samples")
print(f"Test set: {test_df.shape[0]} samples")

# Save the splits
train_df.to_csv('train_data.csv', index=False)
val_df.to_csv('val_data.csv', index=False)
test_df.to_csv('test_data.csv', index=False)

print("\nData preprocessing complete.")

# Step 6: Create Custom Dataset and DataLoaders

# Create PyTorch Dataset
class CityTemperatureDataset(Dataset):
    def __init__(self, dataframe, tokenizer, max_length=128, mode='avg_temp'):
        """
        Custom PyTorch Dataset for city temperature data.
        
        Args:
            dataframe: Pandas DataFrame containing the temperature data
            tokenizer: The tokenizer to use for text encoding
            max_length: Maximum sequence length for tokenization
            mode: Which temperature to predict - 'avg_temp', 'high_temp', or 'low_temp'
        """
        self.dataframe = dataframe
        self.tokenizer = tokenizer
        self.max_length = max_length
        self.mode = mode
        
        # Features to use (combined one-hot encoded and cyclical features)
        self.feature_cols = [col for col in dataframe.columns if col.startswith('city_') 
                             or col in ['month_sin', 'month_cos']]
    
    def __len__(self):
        return len(self.dataframe)
    
    def __getitem__(self, idx):
        # Get the text input
        text = self.dataframe.iloc[idx]['text']
        
        # Tokenize the text
        encoding = self.tokenizer(
            text,
            max_length=self.max_length,
            padding='max_length',
            truncation=True,
            return_tensors='pt'
        )
        
        # Remove the batch dimension added by the tokenizer
        input_ids = encoding['input_ids'].squeeze()
        attention_mask = encoding['attention_mask'].squeeze()
        
        # Get numerical features
        features = torch.tensor(
            self.dataframe.iloc[idx][self.feature_cols].values, 
            dtype=torch.float
        )
        
        # Get the target temperature (ground truth)
        target = torch.tensor(self.dataframe.iloc[idx][self.mode], dtype=torch.float)
        
        return {
            'input_ids': input_ids,
            'attention_mask': attention_mask,
            'features': features,
            'target': target
        }

# Function to prepare Hugging Face datasets
def prepare_hf_datasets(train_df, val_df, test_df):
    """
    Convert pandas DataFrames to Hugging Face Dataset objects
    """
    # Convert to Hugging Face Datasets
    train_dataset = HFDataset.from_pandas(train_df)
    val_dataset = HFDataset.from_pandas(val_df)
    test_dataset = HFDataset.from_pandas(test_df)
    
    print(f"Created Hugging Face datasets:")
    print(f"Train: {len(train_dataset)} samples")
    print(f"Validation: {len(val_dataset)} samples")
    print(f"Test: {len(test_dataset)} samples")
    
    return train_dataset, val_dataset, test_dataset

# Example of loading the data and creating datasets
def create_dataloaders(train_df, val_df, test_df, tokenizer, batch_size=16):
    """
    Create PyTorch DataLoaders from the datasets
    """
    # Create custom datasets
    train_dataset = CityTemperatureDataset(train_df, tokenizer)
    val_dataset = CityTemperatureDataset(val_df, tokenizer)
    test_dataset = CityTemperatureDataset(test_df, tokenizer)
    
    # Create DataLoaders
    train_loader = DataLoader(train_dataset, batch_size=batch_size, shuffle=True)
    val_loader = DataLoader(val_dataset, batch_size=batch_size)
    test_loader = DataLoader(test_dataset, batch_size=batch_size)
    
    print(f"Created DataLoaders with batch size {batch_size}")
    print(f"Train: {len(train_loader)} batches")
    print(f"Validation: {len(val_loader)} batches")
    print(f"Test: {len(test_loader)} batches")
    
    return train_loader, val_loader, test_loader

# Step 7: Define Model Architecture

import torch
import torch.nn as nn
from transformers import BertModel, BertPreTrainedModel

class TemperaturePredictionModel(nn.Module):
    def __init__(self, bert_model_name="bert-base-uncased", num_features=27):
        """
        Model for temperature prediction using BERT and additional features
        
        Args:
            bert_model_name: Name of the pre-trained BERT model
            num_features: Number of additional input features (e.g., one-hot encoded cities + cyclical months)
        """
        super(TemperaturePredictionModel, self).__init__()
        
        # Load pre-trained BERT
        self.bert = BertModel.from_pretrained(bert_model_name)
        
        # Dimension of BERT output
        bert_dim = self.bert.config.hidden_size  # Usually 768 for bert-base
        
        # Feature processing layers
        self.feature_encoder = nn.Sequential(
            nn.Linear(num_features, 64),
            nn.ReLU(),
            nn.Dropout(0.2),
            nn.Linear(64, 32),
            nn.ReLU()
        )
        
        # Combined processing - combines BERT output with features
        self.combined_processor = nn.Sequential(
            nn.Linear(bert_dim + 32, 256),
            nn.ReLU(),
            nn.Dropout(0.3),
            nn.Linear(256, 64),
            nn.ReLU(),
            nn.Dropout(0.2),
            nn.Linear(64, 1)  # Output is a single temperature value
        )
    
    def forward(self, input_ids, attention_mask, features):
        """
        Forward pass of the model
        
        Args:
            input_ids: Tokenized input text
            attention_mask: Attention mask for BERT
            features: Additional numerical features
        
        Returns:
            Predicted temperature value
        """
        # Get BERT output for the [CLS] token
        bert_output = self.bert(input_ids=input_ids, attention_mask=attention_mask)
        cls_output = bert_output.last_hidden_state[:, 0, :]  # [batch_size, bert_dim]
        
        # Process additional features
        encoded_features = self.feature_encoder(features)  # [batch_size, 32]
        
        # Concatenate BERT output with encoded features
        combined = torch.cat([cls_output, encoded_features], dim=1)  # [batch_size, bert_dim + 32]
        
        # Final prediction
        temperature = self.combined_processor(combined).squeeze(-1)  # [batch_size]
        
        return temperature
'''
# Alternative: Smaller Custom Transformer
class CustomTransformerTemperatureModel(nn.Module):
    def __init__(self, vocab_size, embedding_dim=128, nhead=4, num_layers=2, num_features=27):
        """
        A smaller custom transformer model for temperature prediction
        
        Args:
            vocab_size: Size of the vocabulary
            embedding_dim: Dimension of the embeddings
            nhead: Number of attention heads
            num_layers: Number of transformer encoder layers
            num_features: Number of additional numerical features
        """
        super(CustomTransformerTemperatureModel, self).__init__()
        
        # Text embedding
        self.embedding = nn.Embedding(vocab_size, embedding_dim)
        
        # Positional encoding (simplified)
        self.pos_encoder = nn.Embedding(512, embedding_dim)  # Max sequence length = 512
        
        # Transformer encoder
        encoder_layers = nn.TransformerEncoderLayer(
            d_model=embedding_dim, 
            nhead=nhead,
            dim_feedforward=512,
            dropout=0.1
        )
        self.transformer_encoder = nn.TransformerEncoder(encoder_layers, num_layers=num_layers)
        
        # Feature processing layers
        self.feature_encoder = nn.Sequential(
            nn.Linear(num_features, 64),
            nn.ReLU(),
            nn.Linear(64, 32),
            nn.ReLU()
        )
        
        # Combined processing
        self.combined_processor = nn.Sequential(
            nn.Linear(embedding_dim + 32, 128),
            nn.ReLU(),
            nn.Dropout(0.2),
            nn.Linear(128, 32),
            nn.ReLU(),
            nn.Linear(32, 1)
        )
    
    def forward(self, input_ids, attention_mask, features):
        # Create position indices
        positions = torch.arange(0, input_ids.size(1), device=input_ids.device).unsqueeze(0).expand_as(input_ids)
        
        # Get embeddings
        embedded = self.embedding(input_ids)
        pos_embedded = self.pos_encoder(positions)
        embedded = embedded + pos_embedded
        
        # Prepare attention mask for transformer (1=attend, 0=ignore)
        # Transpose because PyTorch transformer expects [seq_len, batch_size, ...]
        transformer_input = embedded.transpose(0, 1)
        
        # Create key padding mask (True=ignore)
        key_padding_mask = (attention_mask == 0)
        
        # Transformer encoding
        transformer_output = self.transformer_encoder(
            transformer_input, 
            src_key_padding_mask=key_padding_mask
        )
        
        # Use the first token's representation as the sentence representation
        sentence_rep = transformer_output[0, :, :]  # First position, all batches, all dims
        
        # Process additional features
        encoded_features = self.feature_encoder(features)
        
        # Concatenate transformer output with encoded features
        combined = torch.cat([sentence_rep, encoded_features], dim=1)
        
        # Final prediction
        temperature = self.combined_processor(combined).squeeze(-1)
        
        return temperature
'''
# Example usage:

# Determine number of features
num_city_features = len(city_encoder.categories_[0])  # Number of cities (one-hot encoded)
num_month_features = 2  # sin and cos of month
total_features = num_city_features + num_month_features

# Initialize model
model = TemperaturePredictionModel(num_features=total_features)

# For custom smaller model:
# model = CustomTransformerTemperatureModel(
#     vocab_size=tokenizer.vocab_size,
#     num_features=total_features
# )

# Print model summary
print(model)


# Step 8: Create and Train a Custom Tokenizer

from tokenizers import Tokenizer, models, pre_tokenizers, decoders, trainers, processors
from tokenizers.models import WordPiece
from tokenizers.trainers import WordPieceTrainer
from tokenizers.pre_tokenizers import Whitespace
from tokenizers.processors import TemplateProcessing

def create_and_train_tokenizer(dataframe, vocab_size=1000, min_frequency=2):
    """
    Create and train a custom WordPiece tokenizer based on the text in our dataset
    """
    # Initialize a tokenizer with a WordPiece model
    tokenizer = Tokenizer(WordPiece(unk_token="[UNK]"))
    
    # Use pre-tokenizer to split text into words
    tokenizer.pre_tokenizer = Whitespace()
    
    # Initialize the trainer
    trainer = WordPieceTrainer(
        vocab_size=vocab_size,
        special_tokens=["[UNK]", "[CLS]", "[SEP]", "[PAD]", "[MASK]"],
        min_frequency=min_frequency
    )
    
    # Prepare the corpus for training
    corpus = dataframe['text'].tolist()
    
    # Train the tokenizer
    tokenizer.train_from_iterator(corpus, trainer)
    
    # Add post-processing for BERT-style inputs
    tokenizer.post_processor = TemplateProcessing(
        single="[CLS] $A [SEP]",
        pair="[CLS] $A [SEP] $B [SEP]",
        special_tokens=[
            ("[CLS]", tokenizer.token_to_id("[CLS]")),
            ("[SEP]", tokenizer.token_to_id("[SEP]"))
        ],
    )
    
    # Add padding capability
    tokenizer.enable_padding(pad_id=tokenizer.token_to_id("[PAD]"), pad_token="[PAD]")
    
    # Add truncation capability
    tokenizer.enable_truncation(max_length=128)
    
    return tokenizer

def convert_to_huggingface_tokenizer(custom_tokenizer, save_path='./tokenizer'):
    """
    Convert the custom tokenizer to a Hugging Face PreTrainedTokenizerFast
    """
    # Save the custom tokenizer
    custom_tokenizer.save(f"tokenizer.json")
    
    # Create a Hugging Face tokenizer from the saved file
    hf_tokenizer = PreTrainedTokenizerFast(
        tokenizer_file=f"tokenizer.json",
        unk_token="[UNK]",
        pad_token="[PAD]",
        cls_token="[CLS]",
        sep_token="[SEP]",
        mask_token="[MASK]"
    )
    
    # Save the Hugging Face tokenizer
    hf_tokenizer.save_pretrained(save_path)
    
    print(f"Tokenizer saved.")
    return hf_tokenizer

# Example usage:

# Load the training data
train_df = pd.read_csv('train_data.csv')

# Create and train the tokenizer
custom_tokenizer = create_and_train_tokenizer(train_df)

# Convert to Hugging Face format
hf_tokenizer = convert_to_huggingface_tokenizer(custom_tokenizer)

# Test the tokenizer
sample_text = "The temperature in London in July is"
encoded = hf_tokenizer(sample_text)
print(encoded)
print(hf_tokenizer.decode(encoded['input_ids']))

# Tokenizer vocabulary size
print(f"Vocabulary size: {hf_tokenizer.vocab_size}")

'''
# Alternative: Use a pre-trained tokenizer
def use_pretrained_tokenizer(model_name="bert-base-uncased"):
    """
    Use a pre-trained tokenizer instead of training our own
    """
    from transformers import AutoTokenizer
    
    tokenizer = AutoTokenizer.from_pretrained(model_name)
    print(f"Loaded pre-trained tokenizer: {model_name}")
    print(f"Vocabulary size: {tokenizer.vocab_size}")
    
    return tokenizer

'''
#### Example usage (): #####
'''
# Load the preprocessed data 
train_df = pd.read_csv('train_data.csv')
val_df = pd.read_csv('val_data.csv')
test_df = pd.read_csv('test_data.csv')

# Create tokenizer (this will be defined in a later step)
tokenizer = create_tokenizer(train_df)

# Create HF datasets
train_hf, val_hf, test_hf = prepare_hf_datasets(train_df, val_df, test_df)

# Create PyTorch DataLoaders
train_loader, val_loader, test_loader = create_dataloaders(
    train_df, val_df, test_df, tokenizer, batch_size=16
)
'''
#### End of Example usage (): #####

# Step 9 & 10: Training Setup and Execution

import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import DataLoader
import numpy as np
from tqdm.auto import tqdm
import matplotlib.pyplot as plt
from sklearn.metrics import mean_squared_error, mean_absolute_error, r2_score

# Custom training loop
def train_model(model, train_loader, val_loader, num_epochs=10, learning_rate=0.001):
    """
    Train the temperature prediction model
    
    Args:
        model: The model to train
        train_loader: DataLoader for training data
        val_loader: DataLoader for validation data
        num_epochs: Number of training epochs
        learning_rate: Learning rate for optimizer
    
    Returns:
        Trained model and training history
    """
    # Set device
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    print(f"Using device: {device}")
    
    # Move model to device
    model = model.to(device)
    
    # Loss function and optimizer
    criterion = nn.MSELoss()
    optimizer = optim.AdamW(model.parameters(), lr=learning_rate)
    
    # Learning rate scheduler
    scheduler = optim.lr_scheduler.ReduceLROnPlateau(
        optimizer, mode='min', factor=0.5, patience=2, verbose=True
    )
    
    # Training history
    history = {
        'train_loss': [],
        'val_loss': [],
        'val_mae': []
    }
    
    # Best model tracking
    best_val_loss = float('inf')
    best_model_state = None
    
    # Training loop
    for epoch in range(num_epochs):
        # Training phase
        model.train()
        train_losses = []
        
        progress_bar = tqdm(train_loader, desc=f"Epoch {epoch+1}/{num_epochs} [Train]")
        for batch in progress_bar:
            # Move batch to device
            input_ids = batch['input_ids'].to(device)
            attention_mask = batch['attention_mask'].to(device)
            features = batch['features'].to(device)
            targets = batch['target'].to(device)
            
            # Forward pass
            optimizer.zero_grad()
            outputs = model(input_ids, attention_mask, features)
            
            # Calculate loss
            loss = criterion(outputs, targets)
            
            # Backward pass and optimize
            loss.backward()
            optimizer.step()
            
            # Track loss
            train_losses.append(loss.item())
            progress_bar.set_postfix({'loss': np.mean(train_losses[-100:])})
        
        # Calculate average training loss for the epoch
        avg_train_loss = np.mean(train_losses)
      

#### EOF ####