# Step 1
## Install the libraries for the application
## pip install pandas numpy matplotlib scikit-learn torch transformers datasets huggingface_hub seaborn

# Step 2 Import Libraries
## Data Architecture
import pandas as pd
import numpy as np

## Visualization
import matplotlib.pyplot as plt
import seaborn as sns

## Preprocessing and Training
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler, OneHotEncoder
import torch
import torch.nn as nn
from torch.utils.data import Dataset, DataLoader

## HuggingFace
from transformers import PreTrainedTokenizerFast
from transformers import Trainer, TrainingArguments
from datasets import Dataset as HFDataSet
import huggingface_hub

# END OF Base Template for LLM #

# Custom or Additional Libraries
from transformers import AutoModelForSequenceClassification

# Step 3 Generate or Load Data
## Generate Data
### create_csv_data.py

## Load Data
csvFile = 'city_temperature_data.csv'
### Use pandas to create the initial dataframe
df = pd.read_csv(csvFile)
### Get some information about this dataframe
print("=== DataFrame Properties ===")
print("DataFrame Shape: ", df.shape)
print("----------------")
print("DataFrame Info: \n")
print(df.info())
print("----------------")
print("Summary Statistics: \n")
print(df.describe())
print("=== End of DataFrame Properties ===")

# Step 4 Visualize Data
## plotting with matplotlib and seaborn
## Exploratory Data Analysis
''' PLOTS
# Plot temperature distribution
plt.figure(figsize=(12, 6))
sns.histplot(df['avg_temp'], bins=20, kde=True)
plt.title('Distribution of Average Temperatures')
plt.xlabel('Temperature (°C)')
plt.ylabel('Frequency')
plt.grid(True, alpha=0.3)
plt.show()

# Plot temperatures by city (boxplot)
plt.figure(figsize=(15, 8))
sns.boxplot(x='city', y='avg_temp', data=df)
plt.title('Average Temperature by City')
plt.xlabel('City')
plt.ylabel('Average Temperature (°C)')
plt.xticks(rotation=90)
plt.grid(True, alpha=0.3)
plt.tight_layout()
plt.show()

# Plot temperature by month
plt.figure(figsize=(12, 6))
monthly_avg = df.groupby('month')['avg_temp'].mean().reindex(
    ["January", "February", "March", "April", "May", "June", 
     "July", "August", "September", "October", "November", "December"]
)
sns.barplot(x=monthly_avg.index, y=monthly_avg.values)
plt.title('Global Average Temperature by Month')
plt.xlabel('Month')
plt.ylabel('Average Temperature (°C)')
plt.xticks(rotation=45)
plt.grid(True, alpha=0.3)
plt.tight_layout()
plt.show()
'''

# Step 5 Engineering the Data
## Created month to number cycle for encoding
month_to_num = {month: i+1 for i, month in enumerate([
    "January", "February", "March", "April", "May", "June", 
    "July", "August", "September", "October", "November", "December"
])}
df['month_num'] = df['month'].map(month_to_num)

## Break into Seasons
df['month_sin'] = np.sin(2 * np.pi * df['month_num']/12)
df['month_cos'] = np.cos(2 * np.pi * df['month_num']/12)

## Create text series from the dataFrame
df['text'] = "The average temperature in " + df['city'] + " in " + df['month'] + " is " + str(df['avg_temp'])

print("=== Sample Data ===")
print(df.head(2))
print("=== End of Sample Data ===")

## OneHotEncoder to Class to encode the city from pandas.dataFrame
city_encoder = OneHotEncoder(sparse_output=False) # sparse is depreciated

city_encoded = city_encoder.fit_transform(df[["city"]])

city_encoded_df = pd.DataFrame(city_encoded, 
                               columns=[f'city_{city}' for city in city_encoder.categories_[0]]
                               )

print("=== Sample Data Encoded DF ===")
print(city_encoded_df.head(2))
city_encoded_df.to_csv('encoded_city.csv')
print("=== End of Sample Data ===")

## Seasonal sin() cos() Scaler -> transformer
### New instance of the Scaler object in memory
scaler = StandardScaler()
### Create a newly named list object in memory for sin() cos()
scaler_features = ['month_sin','month_cos']

### Use the Scaler object, List object to generate the df
#### append the new features to the df -> transform_df
df[scaler_features] = scaler.fit_transform(df[scaler_features])

## Combine the city_transformed data with the seasonal_transformed data to generate our "preprocessed" dataFrame (columns(transfomer/value))
### df(original) + encoded_df
preprocessed_df = pd.concat([
    df[["text", "city", "month", "month_num", "month_sin", "month_cos", "avg_temp","high_temp", "low_temp"]],  # param 1 df
    city_encoded_df # param 2 city_encoded_df
    ], axis=1) # concat

print("=== Sample Data Proprocessed DF ===")
print(preprocessed_df.head(2))
preprocessed_df.to_csv('preprocessed.csv')
print("=== End of Sample Data ===")

###### CUTOFF POINT -> you have to have the data preprocessed ######
# Step 6 Training and Testing -> preprocessing is done
# use the train_test_split() -> 2 params
train_df, temp_df = train_test_split(preprocessed_df, test_size=0.3,random_state=42)

val_df, test_df = train_test_split(temp_df, test_size=0.5, random_state=42)

print("=== Shape of the test_train dataset ====")
print("Preprocessed DF Shape: ", preprocessed_df.shape)
print("Train DF Shape: ", train_df.shape)
print("Test DF Shape: ", test_df.shape)
print("Validation Values DF Shape: ", val_df.shape)

print("=== Data Samples ===")
print("=== Preprocessed Data Samples ===")
print(preprocessed_df.head(5)) # New York Jan-May
print("=== Train Data Samples ===")
print(train_df.head(5)) # Random
print("=== Test Data Samples ===")
print(test_df.head(5)) # Random

# Save the data 
train_df.to_csv('train_data.csv')
test_df.to_csv('test_data.csv')

print("=== Data Processing Completed ===")

# Step 7 dataframes -> Pytorch(DataSets)<tokenize>

## transformer -> encode(numbers) -> preprocessed<train,test,val>
## tokenizer -> encode(text) -> DataSet<pytorch>

## Create PyTorch Dataset, Python Classes <Level 2 Course>
class CityTemperatureDataSet(Dataset):

    def __init__(self, dataframe, tokenizer, max_length=128, mode='avg_temp'):
        """
        Custom PyTorch Dataset for city temperature data. 
        
        Args:
            dataframe: Pandas dataframe containing the temperature data
            tokenizer: The tokenizer to use for text encoding
            max_length: Maximum sequence length for tokenization (defaults to 128 if not passed)
            mode: Which temperature to predict - 'avg_temp','high_temp','low_temp' (defaults to 'avg_temp' if not passed)
        """
        # property class(self.varName) = parameter passed
        self.dataframe = dataframe
        self.tokenizer = tokenizer
        self.max_length = max_length # default is 128
        self.mode = mode # default is 'avg_temp'

        # Define the features to use in the DataSet
        # DataSets "features" -> tbl<field> sheet<column> csv<header> json<key> dataFrame<series>
        ## combination <OneHotEncoder, Cyclical<cyclical,sin,code>>
        self.feature_cols = [
            # [city_Amsterdam,city_Bangkok,city_Beijing, ...]
            col for col in dataframe.columns if col.startswith('city_') or col in ['month_sin', 'month_cos'] 
        ] # self.features

        def __len__(self):
            return len(self.dataframe) # what is the length of the dataset that was passed "control variable<attr>"
        
        # city_, sin, cos, len
        ## next tokenize <text>
        def __getitem__(self, idx):
            # Get the text<char[a-zA-Z0-9specs] input
            text = self.dataframe.iloc[idx]['text']

            # Tokenize <encode text>
            encoding = self.tokenizer(
                text, # getter -> dataframe passed -> 'text'
                max_length = self.max_length, # 128 default
                padding='max_length', # 128 default
                truncation=True,
                return_tensors = 'pt' # pytorch tensor
            ) # self.tokenizer

            # ['input_ids','attention_mask'['idx','text<encoded>']]

            # batch dimensions -> utility objects build dimension
            ## id, mask -> throw these away
            ### input_ids, attention_mask
            ### squeeze() -> get rid of the batch dimensions
            input_ids = encoding['input_ids'].squeeze()
            attention_mask = encoding['attention_mask'].squeeze()

            # Get numerical features <float>
            # 1[idx] -> #month_num, #sin(), #cos()
            # 1 -> [month] -> January
            features = torch.tensor(self.dataframe.iloc[idx][self.feature_cols].values, dtype=torch.float) # [city_NewYork,city_Miami]

            # Get the target temperatures <float>
            # 1[idx] -> [avg,hig,low] -> 56.2
            target = torch.tensor(self.dataframe.iloc[idx][self.mode].values, dtype=torch.float) # default avg_temp

            # return the new pytorch Dataset
            return {
                'input_ids': input_ids,
                'attention_mask': attention_mask,
                # numerical features
                'features': features,
                # targets
                'target': target
            } # return the Dataset to the class object
        
#### return keyword -> new pytorch DataSet to the Application
#### End of the CitytemperatureDataSet Class
# Alternative
## tensor -> preferred format<ONNX> using python

# Step 8 Functions to prepare Hugging Face Datasets
## tensor -> Hugging Face -> export to multiple formats
## pytorchDataSet -> HFDataset-> ONNX, api, ....
## csv -> df.pandas -> proprocessed_df -> sklearn.test_train_split() -> train_df(torch) -> huggingface_dataset
### try: except (MemoryError err)

def prepare_hf_datasets(train_df, val_df, test_df):
    """
    Converts a dataframe pass to the function to Hugging Face Dataset object
    """

    # Dataset = HFDataset(from.pandas)
    train_dataset = HFDataSet.from_pandas(train_df)
    val_dataset = HFDataSet.from_pandas(val_df)
    test_dataset = HFDataSet.from_pandas(test_df)

    print("=== Sample Hugging Face DataSets ===")
    print(f"Train: {len(train_dataset)} samples")
    print(f"Validation Values: {len(val_dataset)} samples")
    print(f"Test: {len(test_dataset)} samples")

    return train_dataset, val_dataset, test_dataset
    # return (ds1, ds2, ds3)
    # End of function

# Create a DataLoader <pytorch lib> to provide an iterable over the dataset 'aka for x in iterate()'
## train_df, train_dataset, CityClass, tokenizer
### batch=16
### Best practice -> train,val,test

'''
utility class
(train, test, val) -> 
covert (train, test, val) -> List [train, test, val]
(List[0], List[2], List[1]) ->
(train, val, test) -> pass to the function ->
'''
def create_dataloaders(train_df, val_df, test_df, tokenizer, batch_size=16):
    # default for the batch is 16, can be overridden by pasing an arg

    # Create Custom Datasets (tokenizer)
    ## In the prior function, we created Hugging Face DataSets (not tokenized)

    train_dataset = CityTemperatureDataSet(train_df, tokenizer)
    test_dataset = CityTemperatureDataSet(test_df, tokenizer)
    val_dataset = CityTemperatureDataSet(val_df, tokenizer)

    # Create DataLoader
    train_loader = DataLoader(train_dataset,batch_size=batch_size,shuffle=True)
    val_loader = DataLoader(val_dataset, batch_size=batch_size)
    test_loader = DataLoader(test_dataset, batch_size=batch_size)

    print("=== Batch Sizes DataLoader ===")
    print(f'Train: {len(train_loader)} batches')
    print(f'Test: {len(test_loader)} batches')
    print(f'Val: {len(val_loader)} batches')

    return train_loader, val_loader, test_loader

# End Data Loaded Function

##### Cut off point csv data -> data -> dataframe -> dataset<HF> -> dataloader<iteration> #####
'''
firstName = "Bob" # variablename, value
In python, type is implicitly applied "string"

string firstName = "Bob" -> explicitly defining as string
java,C,...
'''
# firstName as string assigned value Bob
### objName, type, value

## row 7 -> colA<firstName> -> Cell A7 = Bob
### objName, value

# Step 9 Data Models using torch and BertModel
### objName, type
### firstName as string

# Modeling using DataSet Class

### Structural ###
# "layers" -> predefined torch.nn
### 3 main layers neural network
# Linear Layer -> single incoming transformer Linear "data"
# Convolutional Layer -> image, video data
# Date Series Sequential "Recurrent Layer" -> datetime, Jan,Feb, Sequence

### Functional ###
# Activation Functions -> determine "predictive" output (based on input)

# Loss Functions
## handle predicted loss to be able to traim accuracy

# Utility Function
# help support the building of the models
## nn.Module -> setup class (base class, forwards for inputs, base functional)

# imports need for model architecture
import torch
import torch.nn as nn
from transformers import BertModel # uncased 'best practice term'

class TemperaturePredictionModel(nn.Module):

    def __init__(self, bert_model_name="bert-base-uncased", num_features=32):

        # num_features -> ONEHOTENCODED, cyclical months, ...
        
        # best pracic from helper method
        super(TemperaturePredictionModel, self).__init__()

        # Load the BERT model (or pre-trained model)
        self.bert = BertModel.from_pretrained(bert_model_name)

        # Dimensional size of Bert Output
        bert_dim = self.bert.config.hidden_size # hundreds (32)

        # Features Layer
        ## Linear, ReLU (Activate Function Linear Unit for Linear), Loss Dropout()

        # Define the object in memory
        ## self.feature* = Utility Functions.nn Sequential (FIFO container class)

        self.feature_encoder = nn.Sequential(
            nn.Linear(num_features, 64), # Layer Type
            nn.ReLU(), # Activation Unit
            nn.Dropout(0.2), # higher confidence data lower Dropout()
            nn.Linear(64, 32),
            nn.ReLU()
         ) # feature_encoder
        
        # Combine BERT outputs with features
        self.combined_processor = nn.Sequential(
            nn.Linear(bert_dim + 32, 256),
            nn.ReLU(),
            nn.Dropout(0.2),
            nn.Linear(256, 64),  # featured_encoders(32) + "additional first layer of this function"
            nn.ReLU(),
            nn.Dropout(0.2),
            nn.Linear(64, 1) # output is a single temperature value
        ) # combined_process

    # Forward pass of the model
    def forward(self, input_ids, attention_mask, features):
        # Get the BERT output from console [CLS] token
        bert_output = self.bert(input_ids=input_ids, attention_mask=attention_mask)
        cls_output = bert_output.last_hidden_state[:,0,:]

        # Process the additional features 
        encoded_features = self.feature_encoder(features)

        # Concate cat() BERT output[CLS] with Process Features
        combined = torch.cat([cls_output,encoded_features], dim=1)

        # Final prediction
        temperature = self.combined_processor(combined).squeeze(-1)

        return temperature

###### Cutoff Point -> We now have all the dataframes, datasets, dataloader, model

# Determine OUR features
num_city_features = len(city_encoder.categories_[0]) # number of cities that ONEHOTENCODER encoded
num_month_features = 2 # sin() cos()
total_features = num_city_features + num_month_features

# initialize our predictor
model = TemperaturePredictionModel(num_features=total_features)

print(model)

#### Cutoff Point   data value / data model  ####

# Step 10 Create and Train a Tokenizer
from tokenizers import Tokenizer, models, pre_tokenizers, decoders, trainers, processors
from tokenizers.models import WordPiece
from tokenizers.trainers import WordPieceTrainer
from tokenizers.pre_tokenizers import Whitespace
from tokenizers.processors import TemplateProcessing

def create_and_train_tokenizer(dataframe, vocab_size=1000, min_frequency=2):
    """
    Create and train a custom WordPiece tokenizer based on the text in our dataset
    """
    # Initialize a tokenizer with a WordPiece model
    tokenizer = Tokenizer(WordPiece(unk_token="[UNK]"))
    
    # Use pre-tokenizer to split text into words
    # The hot dog tasted good -> split 5 words
    tokenizer.pre_tokenizer = Whitespace()
    
    # Initialize the trainer
    trainer = WordPieceTrainer(
        vocab_size=vocab_size,
        special_tokens=["[UNK]", "[CLS]", "[SEP]", "[PAD]", "[MASK]"],
        min_frequency=min_frequency
    )
    
    # Prepare the corpus for training
    # [The, hot, dog, tasted, good]
    corpus = dataframe['text'].tolist()
    
    # Train the tokenizer
    tokenizer.train_from_iterator(corpus, trainer)
    
    # Add post-processing for BERT-style inputs
    tokenizer.post_processor = TemplateProcessing(
        single="[CLS] $A [SEP]",
        pair="[CLS] $A [SEP] $B [SEP]",
        special_tokens=[
            ("[CLS]", tokenizer.token_to_id("[CLS]")),
            ("[SEP]", tokenizer.token_to_id("[SEP]"))
        ],
    )
    
    # Add padding capability
    tokenizer.enable_padding(pad_id=tokenizer.token_to_id("[PAD]"), pad_token="[PAD]")
    
    # Add truncation capability
    tokenizer.enable_truncation(max_length=128)
    
    return tokenizer # what will be used to create the json

def convert_to_huggingface_tokenizer(custom_tokenizer, save_path='./tokenizer'):
    """
    Convert the custom tokenizer to a Hugging Face PreTrainedTokenizerFast
    """
    # Save the custom tokenizer
    custom_tokenizer.save(f"tokenizer.json")
    
    # Create a Hugging Face tokenizer from the saved file
    hf_tokenizer = PreTrainedTokenizerFast(
        tokenizer_file=f"tokenizer.json",
        unk_token="[UNK]",
        pad_token="[PAD]",
        cls_token="[CLS]",
        sep_token="[SEP]",
        mask_token="[MASK]"
    )
    
    # Save the Hugging Face tokenizer
    hf_tokenizer.save_pretrained(save_path)
    
    print(f"Tokenizer saved.")
    return hf_tokenizer

# Example usage:

# Load the training data
train_df = pd.read_csv('train_data.csv')

# Create and train the tokenizer
custom_tokenizer = create_and_train_tokenizer(train_df)

# Convert to Hugging Face format
hf_tokenizer = convert_to_huggingface_tokenizer(custom_tokenizer)

# Test the tokenizer
sample_text = "The temperature in London in July is"
encoded = hf_tokenizer(sample_text)
print(encoded)
print(hf_tokenizer.decode(encoded['input_ids']))

# Tokenizer vocabulary size
print(f"Vocabulary size: {hf_tokenizer.vocab_size}")

# Step 11 Setup and Execution of LLM
import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import DataLoader
import numpy as np
from tqdm.auto import tqdm
import matplotlib.pyplot as plt
from sklearn.metrics import mean_squared_error, mean_absolute_error, r2_score

# Custom training loop
def train_model(model, train_loader, val_loader, num_epochs=10, learning_rate=0.001):
    """
    Train the temperature prediction model
    
    Args:
        model: The model to train
        train_loader: DataLoader for training data
        val_loader: DataLoader for validation data
        num_epochs: Number of training epochs
        learning_rate: Learning rate for optimizer
    
    Returns:
        Trained model and training history
    """
    # Set device
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    print(f"Using device: {device}")
    
    # Move model to device
    model = model.to(device)
    
    # Loss function and optimizer
    criterion = nn.MSELoss()
    optimizer = optim.AdamW(model.parameters(), lr=learning_rate)
    
    # Learning rate scheduler
    scheduler = optim.lr_scheduler.ReduceLROnPlateau(
        optimizer, mode='min', factor=0.5, patience=2, verbose=True
    )
    
    # Training history
    history = {
        'train_loss': [],
        'val_loss': [],
        'val_mae': []
    }
    
    # Best model tracking
    best_val_loss = float('inf')
    best_model_state = None
    
    # Training loop
    for epoch in range(num_epochs):
        # Training phase
        model.train()
        train_losses = []
        
        progress_bar = tqdm(train_loader, desc=f"Epoch {epoch+1}/{num_epochs} [Train]")
        for batch in progress_bar:
            # Move batch to device
            input_ids = batch['input_ids'].to(device)
            attention_mask = batch['attention_mask'].to(device)
            features = batch['features'].to(device)
            targets = batch['target'].to(device)
            
            # Forward pass
            optimizer.zero_grad()
            outputs = model(input_ids, attention_mask, features)
            
            # Calculate loss
            loss = criterion(outputs, targets)
            
            # Backward pass and optimize
            loss.backward()
            optimizer.step()
            
            # Track loss
            train_losses.append(loss.item())
            progress_bar.set_postfix({'loss': np.mean(train_losses[-100:])})
        
        # Calculate average training loss for the epoch
        avg_train_loss = np.mean(train_losses)

##### Cutoff -> root.json file and folder->tokenized.json/config.json

# Upload to Hugging Face

print('Thank you for using the LLM Application Builder')
# End of myLLMAplication.py